//////////////////////////////////////////////////////////////
//     ,         //
//                                             //
//    SOM package                                           //
//    (c) Basegroup 2000,                    //
//////////////////////////////////////////////////////////////

unit DBtoSOM;

interface

uses  windows,messages,TDBStats,SysUtils,Classes,DB,Graphics,Math,Controls,c_means;

const
  CellSizeX   = 20;
  CellSizeY   = 20;
  MaxMapNodes = 200;
  MaxInpDim   = 200;

type
  TGridType = (gtRectangle,gtHexagonal);
  TVisualizeMode = (vmMax,vmAvg);
  TVisualizeType = (vtGrayScale,vtColor);
  TTrainType     = (ttLinear,ttBackProportional);
  TInitType      = (itRandom,itSamples);
  //        
  TSOMFields = record
    FieldName : string;        //    DataSet
    FieldNo   : integer;       //    DataSet
    InputNo   : integer;       //  .  
  end;
  TSOMVisualizer = class;
  { ,     }
  TSOMap = class(TComponent)
  private
    fGridType : TGridType;              //  
    fNormalize: boolean;                //   1 
    fWinner  : integer;                 //  
    fXDim,                              //    X
    fYDim    : integer;                 //    Y
    fNodeCnt : integer;                 //  
    fInpVals : array of double;         //   
    fInpDim  : integer;                 //   
    fW : array of array of double ;     //   
    fAlpha   : double;                  //   
    fNeighborhood : double;             //  
    fCorrection : double;               //   
    //    
    function  fGetXWin:integer;
    function  fGetYWin:integer;
    procedure fSetXDim(NewValue:integer);
    procedure fSetYDim(NewValue:integer);
    procedure fSetInpDim(NewValue:integer);
    procedure fSetNormalize(NewValue:boolean);
    procedure fSetGridType(NewValue:TGridType);
    procedure fSetInputs(Index:integer;Value:double);
    function  fGetInputs(Index:integer):double;
    function  fGetNodeX(Index:integer):integer;
    function  fGetNodeY(Index:integer):integer;
  protected
    //  
    procedure CreateSOM ; virtual;
    // c    
    function  CalcCoreDst(n1,n2:integer):integer ; virtual;
    //          n
    function  CalcDist(n:integer):double ; virtual;
    //   
    procedure CorrectNeuron(n:integer) ; virtual;
  public
    //    
    procedure Compute ; virtual;
    //   
    procedure Train   ; virtual;
    //   
    procedure SaveToStream(Stream : TStream)   ; virtual;
    procedure LoadFromStream(Stream : TStream) ; virtual;
    //  -
    property XWin:integer read fGetXWin;
    property YWin:integer read fGetYWin;
    //    
    property NodeCount:integer read fNodeCnt;
    //   - 
    property NWin:integer read fWinner;
    //   Index  
    property NodeX[index:integer]:integer read fGetNodeX;
    property NodeY[index:integer]:integer read fGetNodeY;
    //   
    property Inputs[index:integer]:double read fGetInputs write fSetInputs;
    //      
    property Correction:double read fCorrection;
  published
    //  
    constructor Create(AOwner : TComponent);override;
    destructor Destroy;override;
    //      
    property Normalized:boolean read fNormalize write fSetNormalize;
    //  
    property XDim:integer read fXDim write fSetXDim;
    property YDim:integer read fYDim write fSetYDim;
    //   
    property InputDim:integer read fInpDim write fSetInpDim;
    //   (  
    property GridType:TGridType read fGridType write fSetGridType;
    //  
    property Alpha : double read fAlpha write fAlpha;
    //  
    property Neighborhood : double read fNeighborhood write fNeighborhood;
  end;
  { ,  ,     }
  TDBSOM = class(TComponent)
  private
    fFieldsCount : integer;         //   
    fFields  : array of TSOMFields; //    
    fFldName : TStrings;
    fSOMap   : TSOMap;              //  
    fTrainDS : TDataSet;            //    /
    //  
    fVisMode : TVisualizeMode;            //   
    fHitMap  : array of array of integer; //  
    fHitMax  : integer;                   //   
    fUMatrix : array of double;           // U - 
    fMaxU,fMinU : double;                 //      
    fClusters : array of integer;         //   
    fNClusters : integer;                 //  
    //  
    fInitType         : TInitType;  //   
    fTrainRateInit    : double;     //   
    fTrainRateEnd     : double;     //   
    fDTrainRate       : double;     //   
    fNeighborhoodInit : double;     //   
    fNeighborhoodEnd  : double;     //   
    fDNeighborhood    : double;     //   
    fTrainType        : TTrainType; //    
    fTrainEpoch       : integer;    //    
    fEpochCounter     : integer;    //  
    fEpoch            : integer;    //  
    fCorrection       : double;     //    
    fStopFlag         : boolean;    //   
    fVisList          : TStringList;
    fKeyField         : string;     //    
    fAutoFilter       : boolean;    //  
    fKeyValues : array of array of string;
    fVisSelX          : integer;    //   
    fVisSelY          : integer;    //
    fComputed         : boolean;
    // 
    eAfterTrainStep : TNotifyEvent; //   
    eAfterFiltering : TNotifyEvent; //   
    //  
    function fCalcEuclidianNodeDist(No1,No2:integer):double;
    procedure fAssignFields;
    procedure fTrain;
    procedure fCreateClusters;
    //       
    procedure fSetFields(Value : TStrings);
    function  fGetFields:TStrings;
//    procedure fSetNormalize(Value : boolean);
//    function  fGetNormalize:boolean;
    procedure fAddVisualizer(Value : TSOMVisualizer);
    procedure fDelVisualizer(Value : TSOMVisualizer);
    procedure fVisSel(X,Y:integer);
    procedure fSetVizSelX(Value : integer);
    procedure fSetVizSelY(Value : integer);
    procedure fGetVectorW(WNo : integer;var W : array of double);
    procedure fSetNClusters(Value : integer);
    function  fGetCluster(index : integer):integer;
  protected
    //  ,  
    procedure CreateValuesImage(Field: string;var FreeCells : array of integer;var Values : array of double;var VizMin,VizMax:double) ; virtual;
    //  ,    
    procedure CreateUMatrixValuesImage(var FreeCells : array of integer;var Values : array of double;var VizMin,VizMax:double)        ; virtual;
    //  ,      
    procedure CreateHitMapValuesImage(var FreeCells : array of integer;var Values : array of double;var VizMin,VizMax:double)         ; virtual;
    //  ,  
    procedure CreateClustersImage(var FreeCells : array of integer;var Values : array of double;var VizMin,VizMax:double)             ; virtual;
  public
    //       
    function GetAvailableVisFields:TStrings;
    //   ,      
    function GetAvailableTrainFields:TStrings;
    //       
    procedure FillInputs    ; virtual;
    //  
    procedure Train         ; virtual;
    //   
    procedure CancelTrain   ; virtual;
    //   
    procedure ContinueTrain ; virtual;
    //  
    procedure Compute       ; virtual;
    procedure SaveToStream(S:TStream)  ; virtual;
    procedure LoadFromStream(S:TStream); virtual;
    //      
    property Correction   : double  read fCorrection;
    //      
    property SelX : integer read fVisSelX write fSetVizSelX;
    property SelY : integer read fVisSelY write fSetVizSelY;
    property Cluster[index : integer] : integer read fGetCluster;
  published
    //   
    property SOMap : TSOMap read fSOMap write fSOMap;
    //     
    property DataSet : TDataSet read fTrainDS   write fTrainDS  ;
    //  ,    
    property InputFields  : TStrings read fGetFields write fSetFields;
    //    (     )
    property VisualizeMode: TVisualizeMode read fVisMode write fVisMode;
    //        
    property KeyField     : string read fKeyField write fKeyField;
    //        
    property AutoFilter   : boolean read fAutoFilter write fAutoFilter;
    // ,    
    //   
    property InitializeType : TInitType read fInitType write fInitType;
    // ,    
    //   
    property TrainRateInit : double read fTrainRateInit write fTrainRateInit;
    //   
    property TrainRateEnd  : double read fTrainRateEnd  write fTrainRateEnd;
    //   
    property NeighborhoodInit : double read fNeighborhoodInit write fNeighborhoodInit;
    //   
    property NeighborhoodEnd  : double read fNeighborhoodEnd write fNeighborhoodEnd;
    //    
    property TrainType : TTrainType read fTrainType write fTrainType;
    //     
    property TrainEpoch   : integer read fTrainEpoch  write fTrainEpoch;
    //     
    property EpochCounter : integer read fEpochCounter write fEpochCounter;
    //   
    property Epoch        : integer read fEpoch;
    //    
    property NumberOfClusters : integer read fNClusters write fSetNClusters;
    // 
    //  
    property AfterTrainStep : TNotifyEvent read eAfterTrainStep write eAfterTrainStep;
    //   (  )
    property AfterFiltering : TNotifyEvent read eAfterFiltering write eAfterFiltering;
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
  end;
  {,   }
  TSOMVisualizer = class(TGraphicControl)
  private
    fBitmap : TBitmap;
    fDBSOM  : TDBSOM;
    fField  : string;
    fLabels : array of string; //    
    fOnMouseClick : TNotifyEvent;
    fOnChange     : TNotifyEvent;
    fValues       : array of double;  //    
    fFreeCells    : array of integer; //   
    fVizMax,fVizMin : double;
    fVizType        : TVisualizeType;
    fColorCount     : integer;
    procedure fLeftButtonClick(var Message: TWMMouse );message WM_LBUTTONDOWN;
    procedure fCalcClickCoords(XPos,YPos : integer);
    procedure fCalcCellCenter(X,Y:integer;var CellX,CellY:integer);
    procedure fDrawCell(No:integer);
    procedure fCreateBitmap;  //    
    procedure fDrawSel;
    //    property
    procedure fSetDBSOM(Value : TDBSOM);
    procedure fSetLabels(X,Y:integer;Labels:string);
    function  fGetLabels(X,Y:integer):string;
    function fGetValue : double;
    procedure fSetVizType(Value : TVisualizetype);
    procedure fSetColor(Value : TColor);
  public
    //  
    procedure Paint; override;
    //  
    procedure Compute ; virtual;
    // ,   
    property Labels [X,Y:integer]:string read fGetLabels write fSetLabels;
    // ,   
    property MinValue : double read fVizMin;
    property MaxValue : double read fVizMax;
    // ,    
    property Value : double read fGetValue;
  published
    constructor Create(AOwner : TComponent);override;
    destructor Destroy;override;
    property PopupMenu;
    property Font;
    property Align;
    property Color write fSetColor;
    //  ,   
    property DataField : string read fField write fField;
    //   (    )
    property VisualizeType : TVisualizeType read fVizType write fSetVizType;
    //  
    //  
    property DBSOM : TDBSOM read fDBSOM write fSetDBSOM;
    //     
    property OnMouseClick : TNotifyEvent read fOnMouseClick write fOnMouseClick;
    property OnChange : TNotifyEvent read fOnChange write fOnChange;
  end;

implementation

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//   DBSOM
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
procedure TDBSom.fAddVisualizer;
begin
  // ,       
  if fVisList.IndexOfObject(Value)>0 then exit;
  //    
  fVisList.AddObject(Value.Name,Value);
end;
//    
procedure TDBSom.fSetVizSelX;
begin
  if Value < 0 then Value := 0;
  if Value > fSOMap.XDim-1 then Value := fSOMap.XDim-1;
  fVisSelX := Value;
  fVisSel(fVisSelX,fVisSelY);
end;
procedure TDBSom.fSetVizSelY;
begin
  if Value < 0 then Value := 0;
  if Value > fSOMap.YDim-1 then Value := fSOMap.YDim-1;
  fVisSelY := Value;
  fVisSel(fVisSelX,fVisSelY);
end;
procedure TDBSOM.fSetNClusters;
var
  xI : integer;
begin
  if Value < 1 then
    raise Exception.Create('Number of clusters must be over 0');
  fNClusters := Value;
  if not fComputed then exit;
  fCreateClusters;
  // 
  for xI:=0 to fVisList.Count-1 do
  begin
    with (fVisList.Objects[xI]) as TSOMVisualizer do
    begin
      if DataField <> 'Clusters' then continue;
      Compute;
      Paint;
    end;
  end;
end;
//      
procedure TDBSom.fVisSel;
var
  xI : integer;
  xFilter : string;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  for xI:=0 to fVisList.Count-1 do
  begin
    with (fVisList.Objects[xI]) as TSOMVisualizer do
    begin
      fDBSOM.fVisSelX := X;
      fDBSOM.fVisSelY := Y;
      Paint;
    end;
  end;
  //    ( )
  if not fAutoFilter then exit;
  xFilter := '';
  if Length(fKeyValues) = 0 then exit;
  for xI := 0 to Length(fKeyValues[X+Y*fSOMap.XDim])-1 do
  begin
    if (xI>0) then
      xFilter := xFilter + ' OR ';
    xFilter := xFilter + KeyField+'='''+fKeyValues[X+Y*fSOMap.XDim,xI]+'''';
  end;
  if Length(xFilter) = 0 then xFilter := KeyField + '= 0' + ' AND ' + KeyField +
                                         ' <> 0';
  fTrainDS.DisableControls;
  fTrainDS.Filter := xFilter;
  fTrainDS.Filtered := True;
  fTrainDS.EnableControls;
  if Assigned(eAfterFiltering) then
    eAfterFiltering(Self);
end;
procedure TDBSom.fDelVisualizer;
begin
  // ,       
  if fVisList.IndexOfObject(Value)<0 then exit;
  //  ,  
  fVisList.Delete(fVisList.IndexOfObject(Value));
end;
procedure TDBSOM.fSetFields;
begin
  // c  
  fFldName.Assign(Value);
  fAssignFields;
end;
function TDBSOM.fGetFields;
begin
  Result := fFldName;
end;
{procedure TDBSOM.fSetNormalize;
begin
  fSOMap.Normalized := Value;
end;}
procedure TDBSOM.fGetVectorW;
var
  xI : integer;
begin
  if (WNo<0) or (WNo>=fSOMap.fNodeCnt) then
    raise Exception.Create('Neuron number is out of bounds');
  for xI := 0 to fSOMap.InputDim-1 do
    W[xI] := fSOMap.fW[xI,WNo];
end;
{function TDBSOM.fGetNormalize;
begin
  result := fSOMap.Normalized;
end;}
function TDBSOM.fGetCluster;
begin
  Result := -1;
  if not fComputed then exit;
  if index >= Length(fClusters) then exit;
  Result := fClusters[index];
end;
function TDBSOM.fCalcEuclidianNodeDist(No1,No2:integer):double;
var
  xI : integer;
  xDst : double;
begin
//  if not assigned(fSOMap) then exit;
  if (No1<0) or (No1>=fSOMap.NodeCount) then
    raise Exception.Create(Format('Index No1( %d) out of bounds',[No1]));
  if (No2<0) or (No2>=fSOMap.NodeCount) then
    raise Exception.Create(Format('Index No2( %d) out of bounds',[No2]));
  //  
  xDst := 0;
  for xI := 0 to fSOMap.InputDim-1 do
    xDst := xDst + sqr(fSOMap.fW[xI,No1]-fSOMap.fW[xI,No2]);
  result := sqrt(xDst);
end;

procedure TDBSOM.fAssignFields;
var
  xI : integer;
begin
  SetLength(fFields,fFldName.Count);
  fFieldsCount := fFldName.Count;
  //    
  for xI := 0 to fFldName.Count-1 do
  begin
    fFields[xI].FieldName := fFldName[xI];
    fFields[xI].InputNo   := xI;
  end;
end;

procedure TDBSOM.fTrain;
var
  xI : integer;
  xEpoch : integer;
begin
  //   
  fDTrainRate   := -(fTrainRateInit-fTrainRateEnd)/fTrainEpoch;
  fDNeighborhood:= -(fNeighborhoodInit-fNeighborhoodEnd)/fTrainEpoch;
  //  
  xEpoch := fEpoch;
  for xI:=xEpoch to fEpochCounter-1 do
  begin
    //     
    fTrainDS.First;
    fTrainDS.MoveBy(random(fTrainDS.RecordCount));
    //  
    FillInputs;
    // 
    fSOMap.Compute;
    //   
    if TrainType = ttLinear then
    begin
      fSOMap.Alpha := fTrainRateInit + fDTrainRate*fEpoch;
      if fSOMap.Alpha < fTrainRateEnd then
        fSOMap.Alpha := fTrainRateEnd;
    end
    else
      fSOMap.Alpha := (fTrainRateInit*fSOMap.fNodeCnt)/(fSOMap.fNodeCnt+fEpoch);
    fSOMap.Neighborhood := fNeighborhoodInit + fDNeighborhood*fEpoch;
    if fSOMap.Neighborhood < fNeighborhoodEnd then
      fSOMap.Neighborhood := fNeighborhoodEnd;
    // 
    fSOMap.Train;
    fEpoch := xI;
    fCorrection := fSOMap.Correction;
    if assigned(eAfterTrainStep) then
      eAfterTrainStep(Self);
    if fStopFlag then break;
  end;
end;
procedure TDBSOM.fCreateClusters;
var
  xClusters : TCMeansClustering;
  xI        : integer;
  xW : array of double;
begin
  xClusters := TCMeansClustering.Create(nil);
  xClusters.NumberOfClusters := fNClusters;
  xClusters.NumberOfRecords := fSOMap.fNodeCnt;
  xClusters.DegreeOfRecord := fSOMap.fInpDim;
  xClusters.GetRecordProc := fGetVectorW;
  xClusters.MaxIteration := 1000;
  xClusters.Learn;
  SetLength(xW,fSOMap.fInpDim);
  SetLength(fClusters,SOMap.NodeCount);
  for xI := 0 to fSOMap.fNodeCnt-1 do
  begin
    fGetVectorW(xI,xW);
    fClusters[xI] := xClusters.Analize(xW);
  end;
end;
//   
procedure TDBSOM.FillInputs;
var
  xI : integer;
  xTmp : double;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  for xI := 0 to fFieldsCount-1 do
  begin
    xTmp := fTrainDS.Fields[fFields[xI].FieldNo].AsFloat;
    fSOMap.Inputs[fFields[xI].InputNo] := xTmp;
  end;
end;
//   (     )
procedure TDBSOM.Compute;
var
  xI,xLen : integer;
  xFld    : TField;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  if not assigned(fTrainDS) then
    raise Exception.Create('Visualize dataset is not assigned');
  fTrainDS.DisableControls;
  if fAutoFilter then fTrainDS.Filtered := False;
  fTrainDS.First;
  try
    //   
    for xI:=0 to fFieldsCount-1 do
    begin
      xFld := fTrainDS.FindField(fFields[xI].FieldName);
      if xFld=nil then
        raise Exception.Create(Format('Field %s not found',[fFields[xI].FieldName]));
      fFields[xI].FieldNo := xFld.FieldNo-1;
    end;
    if KeyField <> '' then
    begin
      xFld := fTrainDS.FindField(KeyField);
      if xFld=nil then
      raise Exception.Create(Format('Key field %s not found',[KeyField]));
    end;
    //   
    SetLength(fHitMap,0);
    SetLength(fHitMap,fSOMap.NodeCount);
    SetLength(fKeyValues,0);
    SetLength(fKeyValues,fSOMap.NodeCount);
    fHitMax := 0;
    //   
    fTrainDS.First;
    for xI:=0 to fTrainDS.RecordCount-1 do
    begin
      FillInputs;
      fSOMap.Compute;
      //  
      xLen := Length(fHitMap[fSOMap.NWin]);
      SetLength(fHitMap[fSOMap.NWin],xLen+1);
      if Length(fHitMap[fSOMap.NWin]) > fHitMax then
        fHitMax := Length(fHitMap[fSOMap.NWin]);
      fHitMap[fSOMap.NWin,xLen] := xI;
      //     
      if KeyField <> '' then
      begin
        SetLength(fKeyValues[fSOMap.NWin],xLen+1);
        fKeyValues[fSOMap.NWin,xLen] := fTrainDS.FieldByName(KeyField).AsString;
      end;
      fTrainDS.Next;
    end;
    //    
    SetLength(fUMatrix,fSOMap.fNodeCnt);
    for xI:=0 to fSOMap.fNodeCnt-1 do
    begin
      fUMatrix[xI] := 0;
      //  
      if fSOMap.NodeY[xI]>0 then
      begin
        if (fSOMap.NodeX[xI]>0) and (fSOMap.fGridType = gtRectangle) then
          fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI-fSOMap.XDim-1);
        fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI-fSOMap.XDim);
        if fSOMap.NodeX[xI]<fSOMap.fXDim-1 then
         fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI-fSOMap.XDim+1);
      end;
      if fSOMap.NodeX[xI]>0 then
        fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI-1);
      if fSOMap.NodeX[xI]<fSOMap.fXDim-1 then
        fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI+1);
      if fSOMap.NodeY[xI]<(fSOMap.fYDim-1) then
      begin
        if (fSOMap.NodeX[xI]>0) and (fSOMap.fGridType=gtRectangle) then
          fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI+fSOMap.XDim-1);
        fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI+fSOMap.XDim);
        if fSOMap.NodeX[xI]<fSOMap.fXDim-1 then
         fUMatrix[xI] := fUMatrix[xI]+ fCalcEuclidianNodeDist(xI,xI+fSOMap.XDim+1);
      end;
      if xI = 0 then
      begin
        fMaxU := fUMatrix[0];
        fMinU := fUMatrix[0];
      end;
      if fUMatrix[xI] > fMaxU then
        fMaxU := fUMatrix[xI];
      if fUMatrix[xI] < fMinU then
        fMinU := fUMatrix[xI];
    end;
    fComputed := True;
    fCreateClusters;
    //  
    for xI:=0 to fVisList.Count-1 do
    begin
      with (fVisList.Objects[xI]) as TSOMVisualizer do
      begin
        Compute;
        Paint;
      end;
    end;
  finally
    fTrainDS.EnableControls;
    if fAutoFilter then
    begin
      fVisSel(fVisSelX,fVisSelY);
    end;
  end;
end;
//      
procedure TDBSOM.CancelTrain;
begin
  fStopFlag := True;
end;
//   
procedure TDBSOM.ContinueTrain;
var
  xI   : integer;
  xFld : TField;
  xFiltered : boolean;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  fTrainDS.DisableControls;
  fComputed := False;
  xFiltered := fTrainDS.Filtered;
  if fAutoFilter then fTrainDS.Filtered := False;
  if not assigned(fTrainDS) then
    raise Exception.Create('Train dataset is not assigned');
  try
    fTrainDS.First;
    for xI:=0 to fFieldsCount-1 do
    begin
      xFld := fTrainDS.FindField(fFields[xI].FieldName);
      if xFld=nil then
        raise Exception.Create(Format('Field %s not found',[fFields[xI].FieldName]));
      fFields[xI].FieldNo := xFld.FieldNo-1;
    end;
    //    
    fStopFlag := False;
    fTrain;
  finally
    fTrainDS.EnableControls;
    if fAutoFilter then
      fTrainDS.Filtered := xFiltered;
  end;
end;
//   
procedure TDBSOM.Train;
var
  xI,xJ : integer;
  xFld  : TField;
  xFiltered : boolean;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  if not assigned(fTrainDS) then
    raise Exception.Create('Train dataset is not assigned');
  fComputed := False;
  fTrainDS.DisableControls;
  xFiltered := fTrainDS.Filtered;
  if fAutoFilter then
    fTrainDS.Filtered := False;
  try
    fEpoch := 0;
    //   
    fAssignFields;
    for xI:=0 to fFieldsCount-1 do
    begin
      xFld := fTrainDS.FindField(fFields[xI].FieldName);
      if xFld=nil then
        raise Exception.Create(Format('Field %s not found',[fFields[xI].FieldName]));
      fFields[xI].FieldNo := xFld.FieldNo-1;
    end;
    //    
{    fSOMap.XDim := fMapSizeX;
    fSOMap.YDim := fMapSizeY;       }
    fSOMap.InputDim := fFieldsCount;
    fSOMap.CreateSOM;
    //   
    for xI := 0 to fSOMap.NodeCount-1 do
    begin
      if fInitType = itSamples then
      begin
        fTrainDS.First;
        fTrainDS.MoveBy(Random(fTrainDS.RecordCount));
        for xJ := 0 to fFieldsCount-1 do
        begin
          fSOMap.fW[xJ,xI] := fTrainDS.Fields[fFields[xJ].FieldNo].AsFloat;
        end;
      end;
    end;
    fStopFlag := False;
    fTrain;
  finally
    fTrainDS.EnableControls;
    if fAutoFilter then
      fTrainDS.Filtered := xFiltered;
  end;
end;
//  ,  
procedure TDBSOM.CreateValuesImage;
var
  xI,xJ : integer;
  xLen  : integer;
  xFld  : TField;
  xStats: TDBStatistics;
  xFiltered : boolean;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  if not assigned(fTrainDS) then
    raise Exception.Create('Visualize dataset is not assigned');
  xFld := fTrainDS.FindField(Field);
  if xFld = nil then
    raise Exception.Create(Format('Field %s is not exists',[Field]));
  //   
  fTrainDS.DisableControls;
  xFiltered := fTrainDS.Filtered;
  if fAutoFilter then
    fTrainDS.Filtered := False;
  try
    fTrainDS.First;
    xStats := TDBStatistics.Create(nil);
    xStats.DataSet := fTrainDS;
    xStats.DataField := Field;
    xStats.UpperBound := fTrainDS.RecordCount;
    VizMax := xStats.MaxValue;
    VizMin := xStats.MinValue;
    if fVisMode = vmAvg then
      VizMax := VizMin;
    xStats.Free;
    for xI :=0 to fSOMap.NodeCount-1 do
      Values[xI] := VizMin;
//    fTrainDS.First;
    //  
    for xI:=0 to fSOMap.NodeCount-1 do
    begin
      xLen := length(fHitMap[xI]);
      FreeCells[xI] := integer(xLen <>0);
      if fVisMode = vmMax then
        Values[xI] := VizMin
      else
        Values[xI] := 0;
      for xJ := 0 to xLen-1 do
      begin
        // 
        fTrainDS.First;
        fTrainDS.MoveBy(fHitMap[xI,xJ]);
        //   
        if fVisMode = vmMax then
        begin
          if Values[xI]<xFld.Value then
            Values[xI] := xFld.Value
        end
        else
          Values[xI] := Values[xI]+xFld.Value;
      end;
      if fVisMode=vmAvg then
      begin
        if xLen>0 then
          Values[xI] := Values[xI]/xLen
        else
          Values[xI] := VizMin;
        if Values[xI]>VizMax then
          VizMax := Values[xI];
        if Values[xI]<VizMin then
          VizMin := Values[xI];
      end;
    end;
  finally
    if fAutoFilter then
      fTrainDS.Filtered := xFiltered;
    fTrainDS.EnableControls;
  end;
end;
//  ,       
procedure TDBSOM.CreateUMatrixValuesImage;
var
  xI : integer;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  //  
  VizMax := fMaxU;
  VizMin := fMinU;
  for xI:=0 to fSOMap.NodeCount-1 do
  begin
    Values[xI] := fUMatrix[xI];
    FreeCells[xI] := 1;//integer(Length(fHitMap[xI])<>0);
  end;
end;
//  ,   
procedure TDBSom.CreateClustersImage;
var
  xI : integer;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  for xI := 0 to fSOMap.fNodeCnt-1 do
  begin
    Values[xI] := fClusters[xI];
    FreeCells[xI] := 1;
  end;
  VizMin := 0;
  VizMax := fNClusters-1;
end;
//  ,   
procedure TDBSom.CreateHitMapValuesImage;
var
  xI : integer;
begin
  if not Assigned(fSOMap) then
    raise Exception.Create('SOM is not assigned');
  //  
  VizMax := fHitMax;
  VizMin := 0;
  for xI := 0 to fSOMap.NodeCount-1 do
  begin
    Values[xI] := Length(fHitMap[xI]);
    FreeCells[xI] := integer(Length(fHitMap[xI])<>0);
  end;
end;

function TDBSom.GetAvailableVisFields;
var
  xI : integer;
begin
  result := TStringList.Create;
  result.Clear;
  for xI:=0 to fTrainDS.FieldCount-1 do
    if fTrainDS.Fields[xI].DataType in [ftSmallint,ftInteger,ftWord,ftBoolean,
       ftFloat,ftCurrency,ftAutoInc,ftLargeint] then
      result.Add(fTrainDS.Fields[xI].FieldName);
end;
function TDBSom.GetAvailableTrainFields;
var
  xI : integer;
begin
  if not assigned(fTrainDS) then
    raise Exception.Create('Train DataSet is not assigned');
  result := TStringList.Create;
  result.Clear;
  for xI := 0 to fTrainDS.FieldCount-1 do
    if fTrainDS.Fields[xI].DataType in [ftSmallint,ftInteger,ftWord,ftBoolean,
       ftFloat,ftCurrency,ftAutoInc,ftLargeint] then
      result.Add(fTrainDS.Fields[xI].FieldName);
end;
procedure TDBSOM.SaveToStream;
var
  xTmp : integer;
begin
  //  
  S.Write(fAutoFilter,sizeof(fAutoFilter));
  S.Write(fDNeighborhood,sizeof(fDNeighborhood));
  S.Write(fDTrainRate,sizeof(fDTrainRate));
  S.Write(fEpoch,sizeof(fEpoch));
  S.Write(fEpochCounter,sizeof(fEpochCounter));
  fFldName.SaveToStream(S);
//  S.Write(fGridType,sizeof(fGridType));
  xTmp := Length(fKeyField);
  S.Write(xTmp,sizeof(xTmp));
  S.Write(fKeyField,xTmp);
//  S.Write(fMapSizeX,sizeof(fMapSizeX));
//  S.Write(fMapSizeY,sizeof(fMapSizeY));
  S.Write(fNeighborhoodEnd,sizeof(fNeighborhoodEnd));
  S.Write(fNeighborhoodInit,sizeof(fNeighborhoodInit));
  S.Write(fTrainEpoch,sizeof(fTrainEpoch));
  S.Write(fTrainRateEnd,sizeof(fTrainRateEnd));
  S.Write(fTrainRateInit,sizeof(fTrainRateInit));
  S.Write(fTrainType,sizeof(fTrainType));
  S.Write(fVisMode,sizeof(fVisMode));
  S.Write(fVisSelX,sizeof(fVisSelX));
  S.Write(fVisSelY,sizeof(fVisSelY));
  //  
//  fSOMap.SaveToStream(S);
end;
procedure TDBSOM.LoadFromStream;
var
  xTmp : integer;
begin
  //  
  S.Read(fAutoFilter,sizeof(fAutoFilter));
  S.Read(fDNeighborhood,sizeof(fDNeighborhood));
  S.Read(fDTrainRate,sizeof(fDTrainRate));
  S.Read(fEpoch,sizeof(fEpoch));
  S.Read(fEpochCounter,sizeof(fEpochCounter));
  fFldName.LoadFromStream(S);
//  S.Read(fGridType,sizeof(fGridType));
  S.Read(xTmp,SizeOf(xTmp));
  SetLength(fKeyField,xTmp);
  S.Read(fKeyField,sizeof(fKeyField));
//  S.Read(fMapSizeX,sizeof(fMapSizeX));
//  S.Read(fMapSizeY,sizeof(fMapSizeY));
  S.Read(fNeighborhoodEnd,sizeof(fNeighborhoodEnd));
  S.Read(fNeighborhoodInit,sizeof(fNeighborhoodInit));
  S.Read(fTrainEpoch,sizeof(fTrainEpoch));
  S.Read(fTrainRateEnd,sizeof(fTrainRateEnd));
  S.Read(fTrainRateInit,sizeof(fTrainRateInit));
  S.Read(fTrainType,sizeof(fTrainType));
  S.Read(fVisMode,sizeof(fVisMode));
  S.Read(fVisSelX,sizeof(fVisSelX));
  S.Read(fVisSelY,sizeof(fVisSelY));
  //  
//  fSOMap.LoadFromStream(S);
end;
constructor TDBSOM.Create;
begin
  inherited;
  //   
//  fSOMap := TSOMap.Create(nil);
  fFldName := TStringList.Create;
  fVisList := TStringList.Create;
//  MapSizeX := 1;
//  MapSizeY := 1;
  NeighborhoodInit := 10;
  NeighborhoodEnd  := 0.2;
  TrainRateInit    := 0.2;
  TrainRateEnd     := 0.05;
  EpochCounter     := 10000;
  TrainEpoch       := 5000;
  NumberOfClusters := 10;
end;
destructor TDBSOM.Destroy;
begin
  fFldName.Free;
  SetLength(fFields,0);
  fSOMap.Free;
  inherited;
end;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//                   TSOMap                     !!
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//                   PRIVATE                    !!
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
function  TSOMap.fGetXWin;
begin
  result := fWinner mod fXDim;
end;
function  TSOMap.fGetYWin;
begin
  result := fWinner div fXDim;
end;
procedure TSOMap.fSetXDim(NewValue:integer);
begin
  //  
  if (NewValue<=0) or (NewValue>MaxMapNodes) then
    raise Exception.Create(Format('XDim value %d out of bounds',[NewValue]));
  fXDim := NewValue;
  CreateSOM;
end;
procedure TSOMap.fSetYDim(NewValue:integer);
begin
  if (NewValue<=0) or (NewValue>MaxMapNodes) then
    raise Exception.Create(Format('YDim value %d out of bounds',[NewValue]));
  fYDim := NewValue;
  CreateSOM;
end;
procedure TSOMap.fSetInpDim(NewValue:integer);
begin
  if (NewValue<=0) or (NewValue>MaxInpDim) then
    raise Exception.Create(Format('InpDim value %d out of bounds',[NewValue]));
  fInpDim := NewValue;
  CreateSOM;
end;
procedure TSOMap.fSetNormalize(NewValue:boolean);
begin
  fNormalize := NewValue;
  CreateSOM;
end;
procedure TSOMap.fSetGridType(NewValue:TGridType);
begin
  fGridType := NewValue;
  CreateSOM;
end;
procedure TSOMap.fSetInputs;
begin
  if (Index<0) or (Index>=fInpDim) then
    raise Exception.Create(Format('Input index %d out of bounds',[Index]));
  fInpVals[Index] := Value;
end;
function TSOMap.fGetInputs;
begin
  if (Index<0) or (Index>=fInpDim) then
    raise Exception.Create(Format('Input index %d out of bounds',[Index]));
  result := fInpVals[Index];
end;
function TSOMap.fGetNodeX;
begin
  if (Index<0) or (Index>=fNodeCnt) then
    raise Exception.Create(Format('Node index %d out of bounds',[Index]));
  result := index mod fXDim;
end;
function TSOMap.fGetNodeY;
begin
  if (Index<0) or (Index>=fNodeCnt) then
    raise Exception.Create(Format('Node index %d out of bounds',[Index]));
  result := index div fXDim;
end;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//                   PROTECTED                  !!
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//     
function TSOMap.CalcCoreDst;
var
  xTmp   : integer;
  xL,xR  : integer;
  xYDist : integer;
  x1,y1,x2,y2 : integer;
begin
  //  
  if (n1 >= fNodeCnt) or (n1 < 0) then
    raise Exception.Create(Format('Node index %d out of bounds',[n1]));
  if (n2 >= fNodeCnt) or (n2 < 0) then
    raise Exception.Create(Format('Node index %d out of bounds',[n2]));
  //   
  if n1 > n2 then
  begin // n1 -    
    xTmp := n1; n1:= n2; n2 := xTmp;
  end;
  x1 := n1 mod fXDim;
  x2 := n2 mod fXDim;
  y1 := n1 div fXDim;
  y2 := n2 div fXDim;
  if fGridType = gtRectangle then //  
  begin
    result := Max(abs(x1-x2),abs(y1-y2));
    exit;
  end;
  //  
  //   
  xYDist := (y2-y1);
  xL := x2 - xYDist div 2;
  if ((y2 mod 2)=0) and ((y1 mod 2)=1) then
    dec(xL);
  //     
  if x1<xL then
  begin
    result := xYDist + xL - x1;
    exit;
  end;
  //     
  xR := xL + xYDist;
  if x1>xR then
  begin
    result := xYDist + x1 - xR;
    exit;
  end;
  //  
  result := xYDist;
end;
//         
function TSOMap.CalcDist;
var
  xI   : integer;
  xDst : double;
begin
  xDst := 0;
  for xI:=0 to fInpDim-1 do
    xDst := xDst + sqr(fInpVals[xI]-fW[xI,n]);
  result := xDst;
end;
procedure TSOMAP.CorrectNeuron;
var
  xI:integer;
  xDist  : integer;
  xDelta : double;
begin
  //  
  if fNeighborhood <0 then exit;
  xDist := CalcCoreDst(fWinner,n);
  if (xDist > fNeighborhood*3)and (fNeighborhood>0) then exit;
  for xI:=0 to InputDim-1 do
  begin
    if (fNeighborhood = 0) and (xDist = 0) then
      xDelta := fAlpha*(fInpVals[xI]-fW[xI,n])
    else
      xDelta := fAlpha*exp(-sqr(xDist/fNeighborhood))*(fInpVals[xI]-fW[xI,n]);
    fW[xI,n] := fW[xI,n]+ xDelta;
    if fCorrection < abs(xDelta) then
       fCorrection := abs(xDelta);
  end;
end;
//  
procedure TSOMap.CreateSOM;
var
  xI,xJ : integer;
  xSum  : double;
begin
  //  
  SetLength(fInpVals,fInpDim);
  //     
  fNodeCnt := fXDim*fYDim;
  SetLength(fW,fInpDim);
  xSum := 0;
  for xI :=0 to fInpDim-1 do
  begin
    SetLength(fW[xI],fNodeCnt);
    for xJ:=0 to fNodeCnt-1 do
    begin
      fW[xI,xJ] := Random*0.1-0.05;
      xSum := xSum + sqr(fW[xI,xJ]);
    end;
  end;
  xSum := sqrt(xSum);
  //    
  if fNormalize then
    for xI :=0 to fInpDim-1 do
      for xJ:=0 to fNodeCnt-1 do
        fW[xI,xJ] := fW[xI,xJ]/xSum;
end;
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//                   PUBLIC                     !!
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
procedure TSOMap.Compute;
var
  xI: integer;
  xMin,xTmp : double;
begin
  //    
  if fNormalize then
  begin
    xTmp := 0;
    for xI:=0 to fInpDim-1 do
      xTmp := xTmp + sqr(fInpVals[xI]);
    xTmp := sqrt(xTmp);
    if xTmp <> 0 then
      for xI:=0 to fInpDim-1 do
        fInpVals[xI] := fInpVals[xI]/xTmp;
  end;
  xMin := CalcDist(0);
  fWinner := 0;
  for xI:=1 to fNodeCnt-1 do
  begin
    xTmp := CalcDist(xI);
    if xTmp < xMin then
    begin
      xMin := xTmp;
      fWinner := xI;
    end;
  end;
end;
procedure TSOMap.Train;
var
  xI:integer;
begin
  fCorrection := 0;
  for xI:=0 to fNodeCnt-1 do
  begin
    //   
    CorrectNeuron(xI);
  end;
end;
procedure TSOMap.SaveToStream;
var
  xI,xJ : integer;
begin
  //    
  with Stream do
  begin
    Write(fAlpha,sizeof(fAlpha));
    Write(fCorrection,sizeof(fCorrection));
    Write(fGridType,sizeof(fGridType));
    Write(fInpDim,sizeof(fInpDim));
    Write(fNeighborhood,sizeof(fNeighborhood));
    Write(fNodeCnt,sizeof(fNodeCnt));
    Write(fNormalize,sizeof(fNormalize));
    Write(fXDim,sizeof(fXDIM));
    Write(fYDim,sizeof(fYDIM));
    for xI := 0 to fInpDim-1 do
      for xJ := 0 to fNodeCnt-1 do
        Write(fW[xI,xJ],Sizeof(fW[xI,xJ]));
  end;
end;
procedure TSOMap.LoadFromStream;
var
  xI,xJ : integer;
begin
  //    
  with Stream do
  begin
    Read(fAlpha,sizeof(fAlpha));
    Read(fCorrection,sizeof(fCorrection));
    Read(fGridType,sizeof(fGridType));
    Read(fInpDim,sizeof(fInpDim));
    Read(fNeighborhood,sizeof(fNeighborhood));
    Read(fNodeCnt,sizeof(fNodeCnt));
    Read(fNormalize,sizeof(fNormalize));
    Read(fXDim,sizeof(fXDIM));
    Read(fYDim,sizeof(fYDIM));
    CreateSOM;
    for xI := 0 to fInpDim-1 do
      for xJ := 0 to fNodeCnt-1 do
        Read(fW[xI,xJ],Sizeof(fW[xI,xJ]));
  end;
end;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//                   PUBLISHED                  !!
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
constructor TSOMap.Create;
begin
  inherited;
  fInpDim   := 1;
  fXDim     := 20;
  fYDim     := 20;
  CreateSOM;
end;
destructor TSOMap.Destroy;
begin
  inherited;
end;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//                 TSomVisualizer               !!
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//    
procedure TSOMVisualizer.fLeftButtonClick;
begin
  //   
  if (Message.Keys = MK_LBUTTON) then
    fCalcClickCoords(Message.XPos,Message.YPos);
  inherited;
  if ((Message.Keys and MK_LBUTTON)>0) or
     ((Message.Keys and MK_MBUTTON)>0) or
     ((Message.Keys and MK_RBUTTON)>0)then
    if Assigned(fOnMouseClick) then fOnMouseClick(Self);
end;
//       
procedure TSOMVisualizer.fCalcClickCoords;
var
  xCellWidth,
  xCellHeight : double;
begin
  if not Assigned(fDBSOM) then exit;
  if not assigned(fDBSOM.fSOMap) then exit;
  if fDBSOM.fSOMap.GridType = gtRectangle then
  begin
    //   
    xCellWidth := Width/fDBSOM.fSOMap.XDim;
    xCellHeight := Height/fDBSOM.fSOMap.YDim;
    //    
    fDBSOM.fVisSelX := round(XPos/xCellWidth-0.5);
    fDBSOM.fVisSelY := round(YPos/xCellHeight-0.5);
    // 
  end;
  if fDBSOM.fSOMap.fGridType = gtHexagonal then
  begin
    //   
    xCellWidth := Width/(fDBSOM.fSOMap.XDim+0.5);
    xCellHeight := Height/fDBSOM.fSOMap.YDim;
    //    
    fDBSOM.fVisSelY := round(YPos/xCellHeight-0.5);
    if (fDBSOM.fVisSelY mod 2) = 0 then
      fDBSOM.fVisSelX := round(XPos/xCellWidth-0.5)
    else
      fDBSOM.fVisSelX := round((XPos-xCellWidth/2)/xCellWidth-0.5);
    // 
  end;
  if fDBSOM.fVisSelX > (fDBSOM.fSOMap.fXDim-1) then
    fDBSOM.fVisSelX := fDBSOM.fSOMap.fXDim-1;
  if fDBSOM.fVisSelX < 0 then
    fDBSOM.fVisSelX := 0;
  if fDBSOM.fVisSelY > (fDBSOM.fSOMap.fYDim-1) then
    fDBSOM.fVisSelY := fDBSOM.fSOMap.fYDim-1;
  if fDBSOM.fVisSelY < 0 then
    fDBSOM.fVisSelY := 0;
  fDBSOM.fVisSel(fDBSOM.fVisSelX,fDBSOM.fVisSelY);
end;
//   
procedure TSOMVisualizer.fSetDBSOM;
begin
//  if not Assigned(DBSOM) then exit;
//  if not Assigned(DBSOM.fSOMap) then exit;
  if Assigned(fDBSOM) then
    fDBSOM.fDelVisualizer(Self);
  fDBSOM := Value;
  fDBSOM.fAddVisualizer(Self);
end;
procedure TSOMVisualizer.fSetVizType;
begin
  fVizType := Value;
  fCreateBitmap;
  Paint;
end;
procedure TSOMVisualizer.fCalcCellCenter;
var
  xCellWidth,
  xCellHeight : double;
begin
  xCellWidth := Width/fDBSOM.fSOMap.XDim;
  xCellHeight := Height/fDBSOM.fSOMap.YDim;
  CellY := round((Y+0.5)*xCellHeight);
  if fDBSOM.fSOMap.fGridType = gtRectangle then
  begin
    CellX := round((X+0.5)*xCellWidth);
  end;
  if fDBSOM.fSOMap.fGridType = gtHexagonal then
  begin
    xCellWidth :=(Width)/(fDBSOM.fSOMap.XDim+0.5);
    CellX := round((X+0.5)*xCellWidth);
    if (Y mod 2)=1 then
      CellX := CellX + round(xCellWidth/2);
  end;
end;
function TSOMVisualizer.fGetValue;
begin
  if not Assigned(DBSOM) then exit;
  if not Assigned(DBSOM.fSOMap) then exit;
  if Length(fValues) = 0 then Result := 0
  else result := fValues[fDBSOM.fVisSelX+fDBSOM.fVisSelY*fDBSOM.fSOMap.fXDim];
end;
procedure TSOMVisualizer.fDrawCell;
var
  x0,y0 : integer;
  xClr  : double;
  xR,xG,xB : double;
  xPoints : array[0..5] of TPoint;
  xD      : integer;
begin
  xClr := fValues[No];
  //   
  if fFreeCells[No]>0 then
  begin
    if fVizMax<> fVizMin then
    begin
      xClr := xClr - fVizMin; //    [0..fVizMax-fVizMin]
      xClr := (xClr)/(fVizMax-fVizMin);
    end;
    xClr := round(xClr * fColorCount);
    xClr := xClr / fColorCount;
    if xClr < 0 then xClr := 0;
    if xClr > 1 then xClr := 1;
    //     
    if fVizType = vtGrayScale then
    begin
      xClr := 1 - xClr;
      begin
        xR := xClr*254;
        xG := xClr*254;
        xB := xClr*254;
      end
    end
    else
    begin
      xR := (xClr)*254;
      if (xClr>=0) and (xClr<=0.5) then
        xG := (xClr+0.5)*254
      else
        if (xClr>=0.5) and (xClr<=1) then
          xG := (1-xClr+0.5)*254
        else
          xG := 0;
      xB := (1-xClr)*254;
    end;
    fBitmap.Canvas.Brush.Color := RGB(round(xR),round(xG),round(xB));
    fBitmap.Canvas.Pen.Color := RGB(round(xR),round(xG),round(xB));
  end
  else
  begin
    fBitmap.Canvas.Brush.Color := Color;;
    fBitmap.Canvas.Pen.Color := Color;
  end;
  //   
  if fDBSOM.fSOMap.fGridType = gtRectangle then
  begin
    //   
    x0 := (No mod fDBSOM.fSOMap.fXDim)*CellSizeX;
    y0 := (No div fDBSOM.fSOMap.fXDim)*CellSizeY;
    //  
    fBitmap.Canvas.FillRect(Rect(round(x0),round(y0),round(x0+CellSizeX),
                            round(y0+CellSizeY)));
  end;
  if fDBSOM.fSOMap.fGridType = gtHexagonal then
  begin
    xD := (CellSizeY div 4)*3;
    //   
    x0 := fDBSOM.fSOMap.NodeX[No];
    y0 := fDBSOM.fSOMap.NodeY[No];
    //  ?
    if (round(y0) mod 2)=0 then
      x0 := x0*CellSizeX
    else
      x0 := x0*CellSizeX + (CellSizeX div 2);
    y0 := round(y0*xD);
    xPoints[0].x := round(x0+CellSizeX div 2);
    xPoints[0].y := round(y0);
    xPoints[1].x := round(x0+CellSizeX);
    xPoints[1].y := round(y0+CellSizeY div 4);
    xPoints[2].x := round(x0+CellSizeX);
    xPoints[2].y := round(y0+xD);
    xPoints[3].x := round(x0+CellSizeX div 2);
    xPoints[3].y := round(y0+CellSizeY);
    xPoints[4].x := round(x0);
    xPoints[4].y := round(y0+xD);
    xPoints[5].x := round(x0);
    xPoints[5].y := round(y0+CellSizeY div 4);
    fBitmap.Canvas.Polygon(xPoints);
  end;
end;

procedure TSOMVisualizer.fCreateBitmap;
var
  xI : integer;
begin
  //   
  fBitmap.Canvas.Brush.Color := Color;//RGB(0,0,255);
  fBitmap.Canvas.FillRect(Rect(0,0,fBitmap.Width,fBitmap.Height));
  if not assigned(fDBSOM) then exit;
  if not assigned(fDBSOM.fSOMap) then exit;
  if Length(fValues)<DBSOM.SOMap.NodeCount then exit;
  if fBitmap = nil then exit;
  if fDBSOM.fSOMap.fGridType = gtRectangle then
  begin
    fBitmap.Height := fDBSOM.fSOMap.YDim*CellSizeY;
    fBitmap.Width  := fDBSOM.fSOMap.XDim*CellSizeX;
  end
  else
  begin
    fBitmap.Height := round(fDBSOM.fSOMap.YDim * (CellSizeY div 4)*3 + CellSizeY div 4);
    fBitmap.Width  := fDBSOM.fSOMap.XDim*CellSizeX+CellSizeX div 2;
  end;
  //     
  for xI:=0 to fDBSOM.fSOMap.NodeCount-1 do
    fDrawCell(xI);
end;

procedure TSOMVisualizer.fDrawSel;
var
  xX,xY,xI,xC,yC : integer;
begin
  if not assigned(fDBSom) then exit;
  if not assigned(fDBSom.fSOMap) then exit;
  xX := 0;
  //      
  fCalcCellCenter(fDBSOM.fVisSelX,fDBSOM.fVisSelY,xX,xY);
  //  
  Canvas.Pen.Color := clWhite;
  Canvas.Brush.Style := bsClear;
  Canvas.Ellipse(xX-3,xY-3,xX+3,xY+3);
  Canvas.Pen.Color := clBlack;
  Canvas.Ellipse(xX-2,xY-2,xX+2,xY+2);
  Canvas.Ellipse(xX-4,xY-4,xX+4,xY+4);
  //  
  Canvas.Font := Font;
  for xI:=0 to fDBSOM.fSOMap.fXDim*fDBSOM.fSOMap.fYDim-1 do
  begin
    if xI >= Length(fLabels) then break;
    if fLabels[xI] = '' then continue;
    //   
    xC := xI mod fDBSOM.fSOMap.fXDim;
    yC := xI div fDBSOM.fSOMap.fXDim;
    fCalcCellCenter(xC,yC,xX,xY);
    Canvas.TextRect(Rect(0,0,Width,Height),xX,xY,fLabels[xI])
  end;
  if Assigned(fOnChange) then fOnChange(Self);
end;
procedure TSOMVisualizer.fSetLabels;
begin
  if not Assigned(DBSOM) then exit;
  if not Assigned(DBSOM.fSOMap) then exit;
  if (X >= DBSOM.fSOMap.XDim) or (X<0) then
    raise Exception.Create('X index is out of bounds');
  if (Y >= DBSOM.fSOMap.YDim) or (Y<0) then
    raise Exception.Create('Y index is out of bounds');
  //  
  fLabels[X+Y*fDBSOM.fSOMap.XDim] := Labels;
  // 
  Paint;
end;
function  TSOMVisualizer.fGetLabels;
begin
  if not Assigned(DBSOM) then exit;
  if not Assigned(DBSOM.fSOMap) then exit;
  if (X >= DBSOM.fSOMap.XDim) or (X<0) then
    raise Exception.Create('X index is out of bounds');
  if (Y >= DBSOM.fSOMap.YDim) or (Y<0) then
    raise Exception.Create('X index is out of bounds');
  //  
  Result := fLabels[X+Y*fDBSOM.fSOMap.XDim];
end;
procedure TSomVisualizer.fSetColor(Value : TColor);
begin
  inherited Color := Value;
  fCreateBitmap;
  Paint;
end;


constructor TSomVisualizer.Create;
begin
  inherited;
  fBitmap := TBitmap.Create;
  fBitmap.Height := 100;
  fBitmap.Width := 128;
  Height := 100;
  Width  := 128;
  fColorCount := 200;
  fBitmap.Canvas.Brush.Color := Color;
  fBitmap.Canvas.FillRect(Rect(0,0,fBitmap.Width,fBitmap.Height));
end;
destructor TSomVisualizer.Destroy;
begin
  fBitmap.Free;
  if Assigned(fDBSOM) then
    fDBSOM.fDelVisualizer(Self);
  inherited;
end;
procedure TSomVisualizer.Compute;
begin
  if not assigned(fDBSOM) then exit;
  if not assigned(fDBSOM.fSOMap) then exit;
  if not fDBSOM.fComputed then exit;
  //  
  SetLength(fValues,fDBSOM.fSOMap.fNodeCnt);
  SetLength(fFreeCells,fDBSOM.fSOMap.fNodeCnt);
  if fField = 'UMatrix' then
    fDBSOM.CreateUMatrixValuesImage(fFreeCells,fValues,fVizMin,fVizMax)
  else
    if fField = 'HitMap' then
      DBSOM.CreateHitMapValuesImage(fFreeCells,fValues,fVizMin,fVizMax)
    else
      if fField = 'Clusters' then
      begin
        DBSOM.CreateClustersImage(fFreeCells,fValues,fVizMin,fVizMax);
      end
      else
      begin
        fDBSOM.CreateValuesImage(fField,fFreeCells,fValues,fVizMin,fVizMax);
      end;
  //   
  fCreateBitmap;
  SetLength(fLabels,fDBSOM.fSOMap.fXDim*fDBSOM.fSOMap.YDim);
  Paint;
end;

procedure TSomVisualizer.Paint;
begin
  inherited;
//  if not Assigned(DBSOM) then exit;
//  if not Assigned(DBSOM.fSOMap) then exit;
  Canvas.StretchDraw(ClientRect,fBitmap);
  fDrawSel;
end;

end.
