unit SOMComponentsEditors;

interface

uses
  DesignIntf,  DesignEditors,
  SysUtils, Classes, Controls, DB,
  DBtoSOM, TDBStats,
  fmDBSOMInputFieldsEditDlg;
{*********************************************************************}
{                         }
{                                       }
{*********************************************************************}
type
  TDBStringProperty = class(TStringProperty)
  public
    function GetAttributes : TPropertyAttributes; override;
    procedure GetValueList(List: TStrings); virtual; abstract;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

  TDBStatisticsDataFieldProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

  TSOMVisualizerDataFieldProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings);override;
  end;

  TDBSOMKeyFieldProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings);override;
  end;

  TDBSOMInputFieldsProperty = class(TStringProperty)
  public
    function GetAttributes : TPropertyAttributes; override;
    function GetValue : string; override;
    procedure Edit; override;
  end;

procedure Register;

implementation

{TDBStringProperty}
function TDBStringProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TDBStringProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    GetValueList(Values);
    for I := 0 to Values.Count - 1 do Proc(Values[I]);
  finally
    Values.Free;
  end;
end;

{TDBStatisticsDataFieldProperty}
procedure TDBStatisticsDataFieldProperty.GetValueList(List: TStrings);
var
  DBStats1 : TDBStatistics;
  DataSet : TDataSet;
begin
  DBStats1 := GetComponent(0) as TDBStatistics;
  List.Clear;
  if (DBStats1.DataSet = nil) then exit;
  DataSet := DBStats1.DataSet;
  DataSet.GetFieldNames(List);
end;

{TDBSOMKeyFieldProperty}
procedure TDBSOMKeyFieldProperty.GetValueList(List: TStrings);
var
  DBSOM : TDBSOM;
begin
  DBSOM := GetComponent(0) as TDBSOM;
  if DBSOM = nil then exit;
  List.Clear;
  if DBSOM.DataSet = nil then exit;
  DBSOM.DataSet.GetFieldNames(List);
end;

{TSOMVisualizerDataFieldProperty}
procedure TSOMVisualizerDataFieldProperty.GetValueList(List: TStrings);
var
  DBSOM : TDBSOM;
begin
     DBSOM := (GetComponent(0) as TSOMVisualizer).DBSOM;
     if DBSOM = nil then exit;
     List.Clear;
     if DBSOM.DataSet <> nil then
       List.Assign(DBSOM.GetAvailableVisFields);
     List.Add('HitMap');
     List.Add('UMatrix');
     List.Add('Clusters');
end;

{TDBSOMInputFieldsProperty}
function TDBSOMInputFieldsProperty.GetAttributes;
begin
  Result := [paDialog,paMultiSelect];
end;

function TDBSOMInputFieldsProperty.GetValue;
begin
  Result := Format('(%s)',[GetPropType^.Name]);
end;

procedure TDBSOMInputFieldsProperty.Edit;
var
  xDlg   : TfrmDBSOMInputFieldsEditDlg;
  xDBSOM : TDBSOM;
  xI     : integer;
begin
  xDBSOM := GetComponent(0) as TDBSOM;
  if xDBSOM.DataSet = nil then exit;
  xDlg := TfrmDBSOMInputFieldsEditDlg.Create(nil);
  try
    //     
    xDlg.clbAvailableFields.Items.Assign(xDBSOM.GetAvailableTrainFields);
    for xI := 0 to xDlg.clbAvailableFields.Items.Count-1 do
      if xDBSOM.InputFields.IndexOf(xDlg.clbAvailableFields.Items[xI])>=0 then
        xDlg.clbAvailableFields.Checked[xI] := True
      else
        xDlg.clbAvailableFields.Checked[xI] := False;
      if xDlg.ShowModal = mrOk then
      begin
        xDBSOM.InputFields.Clear;
        for xI := 0 to xDlg.clbAvailableFields.Items.Count-1 do
          if xDlg.clbAvailableFields.Checked[xI] then
            xDBSOM.InputFields.Add(xDlg.clbAvailableFields.Items[xI]);
      end;
  finally
    xDlg.Free;
  end;
end;

procedure Register;
begin
  RegisterComponents('SomKohonen',[TDBSOM,TSOMap,TSOMVisualizer]);
  RegisterPropertyEditor(TypeInfo(String),TSOMVisualizer,'DataField',TSOMVisualizerDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(String),TDBSOM,'KeyField',TDBSOMKeyFieldProperty);
  RegisterPropertyEditor(TypeInfo(TStrings),TDBSOM,'InputFields',TDBSOMInputFieldsProperty);
  RegisterPropertyEditor(TypeInfo(String),TDBStatistics,'DataField',TDBStatisticsDataFieldProperty);
end;

end.
