unit TDBStats;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DesignIntf,{ DesignEditors,}
  DB, DBTables, Math;

const
     MaxValues = 100000;
type
  TDBStatistics = class(TComponent)
  private
    fRecords : LongInt;
    Data: array[1..MaxValues] of Double;
    IsArrayFilled : Boolean;
    DidGetAll: Boolean;
    fMean : Extended;
    fSum : Extended;
    fSumOfSquares : Extended;
    fMinValue : Extended;
    fMaxValue : Extended;
    fStdDev : Extended;
    fPopnStdDev : Extended;
    fVariance : Extended;
    fPopnVariance : Extended;
    fTotalVariance : Extended;
    fNorm : Extended;
    fSkew : Extended;
    fKurtosis : Extended;
    fLowerBound : LongInt;
    fUpperBound : LongInt;
    fDataSet : TDataSet;
    fDataField : String;
    procedure SetDataSet(Value: TDataSet);
    procedure SetDataField(Value : String);
    procedure FillArray;
    function GetMean: Extended;
    function GetSum: Extended;
    function GetSumOfSquares: Extended;
    function GetMinValue: Extended;
    function GetMaxValue: Extended;
    function GetStdDev: Extended;
    function GetPopnStdDev: Extended;
    function GetVariance: Extended;
    function GetPopnVariance: Extended;
    function GetTotalVariance: Extended;
    function GetNorm: Extended;
    function GetSkew: Extended;
    function GetKurtosis: Extended;
    function GetRange: LongInt;    
  public
    procedure GetAllStats;
    property Records : LongInt read fRecords;
    property Mean: Extended read GetMean;
    property Sum: Extended read GetSum;
    property SumOfSquares: Extended read GetSumOfSquares;
    property MinValue: Extended read GetMinValue;
    property MaxValue: Extended read GetMaxValue;
    property StdDev: Extended read GetStdDev;
    property PopnStdDev: Extended read GetPopnStdDev;
    property Variance: Extended read GetVariance;
    property PopnVariance: Extended read GetPopnVariance;
    property TotalVariance : Extended read GetTotalVariance;
    property Norm: Extended read GetNorm;
    property Skew: Extended read GetSkew;
    property Kurtosis : Extended read GetKurtosis;
  published
    property DataSet : TDataSet read fDataSet write SetDataSet;
    property DataField : String read fDataField write SetDataField;
    property LowerBound : LongInt read fLowerBound write fLowerBound;
    property UpperBound : LongInt read fUpperBound write fUpperBound;
  end;

implementation
{df$R *.dcr}
function TDBStatistics.GetRange: LongInt;
var
 xTmp : LongInt;
begin
  if (LowerBound < 1) then LowerBound := 1;
  if (UpperBound < 1) then UpperBound := LowerBound + 1;
  {if they're in the wrong order, swap 'em}
  if (LowerBound > UpperBound) then
  begin
    xTmp := UpperBound;
    UpperBound := LowerBound;
    LowerBound := xTmp;
  end;
  if (LowerBound > Records) then LowerBound := 1;
  if (UpperBound > Records) then UpperBound := Records;
  if (UpperBound - LowerBound > MaxValues) then UpperBound := LowerBound + MaxValues;
  Result := UpperBound - LowerBound;
end;

procedure TDBStatistics.FillArray;
var
 xI, xIndex : LongInt;
 xOpenFlag      : boolean;
begin
  if (fDataSet = nil) then exit;
  xIndex := 1;
  GetRange; {Do error checking!}
  xOpenFlag := False;
  if not DataSet.Active then
  begin
     fDataSet.Open;
     xOpenFlag := True;
  end;
  fDataSet.First;
  fDataSet.MoveBy(LowerBound);
  if ((fDataSet.FieldByName(fDataField) is TCurrencyField) or
     (fDataSet.FieldByName(fDataField) is TFloatField) or
     (fDataSet.FieldByName(fDataField) is TIntegerField) or
     (fDataSet.FieldByName(fDataField) is TSmallIntField))then
  begin
    for xI := LowerBound to UpperBound do
    begin
      Application.ProcessMessages;
      if not (fDataSet.FieldByName(fDataField).IsNull) then
        Data[xIndex] := fDataSet.FieldByName(fDataField).Value
      else
        Data[xIndex] := 0;
      Inc(xIndex);
      fDataSet.Next;
    end;
  end;
  if xOpenFlag then
    fDataSet.Close;
  IsArrayFilled := True;
  DidGetAll := False;
end;

procedure TDBStatistics.GetAllStats;
var
  M1,M2,M3,M4,Sk,Kurt : Extended;
  TempRange : Integer;
begin
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  TempRange := GetRange;
  if (TempRange = 0) then exit;
  fMean := Math.Mean(Slice(Data,TempRange));
  fSum := Math.Sum(Slice(Data,TempRange));
  Application.ProcessMessages;
  fSumOfSquares := Math.SumOfSquares(Slice(Data,TempRange));
  {$IFDEF VER90} {if we're running Delphi2}
     fMinValue := Math.MaxValue(Slice(Data,TempRange));
     fMaxValue := Math.MinValue(Slice(Data,TempRange));
  {$ELSE}
     fMinValue := Math.MinValue(Slice(Data,TempRange));
     fMaxValue := Math.MaxValue(Slice(Data,TempRange));
  {$ENDIF}
   if fMinValue <> fMaxValue then
   begin
     fStdDev := Math.StdDev(Slice(Data,TempRange));
     Application.ProcessMessages;
     fPopnStdDev := Math.PopnStdDev(Slice(Data,TempRange));
     fVariance := Math.Variance(Slice(Data,TempRange));
     fPopnVariance := Math.PopnVariance(Slice(Data,TempRange));
     fTotalVariance := Math.TotalVariance(Slice(Data,TempRange));
     Application.ProcessMessages;
     Math.MomentSkewKurtosis(Slice(Data,TempRange), M1,M2,M3,M4,Sk,Kurt);
     fSkew := Sk;
     fKurtosis := Kurt;
   end
   else
   begin
     fStdDev := 0;
     fPopnStdDev := 0;
     fVariance := 0;
     fPopnVariance := 0;
     fTotalVariance := 0;
     Application.ProcessMessages;
     fSkew := 0;
     fKurtosis := 0;
   end;
   fNorm := Math.Norm(Slice(Data,TempRange));
   DidGetAll := True;
end;

procedure TDBStatistics.SetDataSet(Value: TDataSet);
var
  xOpenFlag : boolean;
begin
  if Value = nil then
  begin
    IsArrayFilled := False;
    DidGetAll := False;
    fDataSet := Value;
    exit;
  end;
  IsArrayFilled := False;
  DidGetAll := False;
  fDataSet := Value;
  xOpenFlag := False;
  {Get the Number of records just once}
  try
    if not fDataSet.Active then
    begin
      fDataSet.Open;
      xOpenFlag := True;
    end;
    fRecords := fDataSet.RecordCount;
    if xOpenFlag then
      fDataSet.Close;
  except
  end;
end;

procedure TDBStatistics.SetDataField(Value : String);
begin
  IsArrayFilled := False;
  DidGetAll := False;
  fDataField := Value;
end;

function TDBStatistics.GetMean: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fMean := Math.Mean(Slice(Data,GetRange));
  Result := fMean;
end;

function TDBStatistics.GetSum: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fSum := Math.Sum(Slice(Data,GetRange));
  Result := fSum;
end;

function TDBStatistics.GetSumOfSquares: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fSumOfSquares := Math.SumOfSquares(Slice(Data,GetRange));
  Result := fSumOfSquares;
end;

function TDBStatistics.GetMinValue: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    {Under Delphi2, actually calls MaxValue due to small bug in Math unit}
    {$IFDEF VER90}
       fMinValue := Math.MaxValue(Slice(Data,GetRange));
    {$ELSE}
       fMinValue := Math.MinValue(Slice(Data,GetRange));
    {$ENDIF}
  Result := fMinValue;
end;

function TDBStatistics.GetMaxValue: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    {actually calls MinValue due to small bug in Math unit}
    {$IFDEF VER90}
       fMaxValue := Math.MinValue(Slice(Data,GetRange));
    {$ELSE}
       fMaxValue := Math.MaxValue(Slice(Data,GetRange));
    {$ENDIF}
  Result := fMaxValue;
end;

function TDBStatistics.GetStdDev: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fStdDev := Math.StdDev(Slice(Data,GetRange));
  Result := fStdDev;
end;

function TDBStatistics.GetPopnStdDev: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fPopnStdDev := Math.PopnStdDev(Slice(Data,GetRange));
  Result := fPopnStdDev;
end;

function TDBStatistics.GetVariance: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fVariance := Math.Variance(Slice(Data,GetRange));
  Result := fVariance;
end;

function TDBStatistics.GetPopnVariance: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fPopnVariance := Math.PopnVariance(Slice(Data,GetRange));
  Result := fVariance;
end;

function TDBStatistics.GetTotalVariance: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fTotalVariance := Math.TotalVariance(Slice(Data,GetRange));
  Result := fTotalVariance;
end;

function TDBStatistics.GetNorm: Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
    fNorm := Math.Norm(Slice(Data,GetRange));
  Result := fNorm;
end;

function TDBStatistics.GetSkew: Extended;
var
   M1,M2,M3,M4,Sk,Kurt : Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
  begin
    Math.MomentSkewKurtosis(Slice(Data,GetRange),M1,M2,M3,M4,Sk,Kurt);
    fSkew := Sk;
  end;
  Result := fSkew;
end;

function TDBStatistics.GetKurtosis: Extended;
var
   M1,M2,M3,M4,Sk,Kurt : Extended;
begin
  Result := 0;
  if (fDataSet = nil) then exit;
  if not (IsArrayFilled) then FillArray;
  if not (DidGetAll) then
  begin
    Math.MomentSkewKurtosis(Slice(Data,GetRange), M1,M2,M3,M4,Sk,Kurt);
    fKurtosis := Kurt;
  end;
  Result := fKurtosis;
end;
end.
