unit c_means;

interface

uses SysUtils,Classes;

type
  GetRecordProc = procedure (RecNo : integer; var Values : array of double) of object;
  TCMeansClustering = class(TComponent)
  private
    fV   : array of array of double; //   
    fW   : array of integer;         //    
    fNOC : integer;  //  
    fNOR : integer;  //    
    fDOR : integer;  //  
    fGetRecordProc : GetRecordProc;
    fAfterEpochProc: TNotifyEvent;
    fDelta : double; //   
    fValues        : array of double;
    fMaxIteration : integer;
    fIteration    : integer;
    fStopCheck    : array [0..1] of array of integer;
    fNum          : integer;
    fClustersW    : array of integer;
    //  
    procedure fInit;
    //    
    procedure fSetNOC(Value : integer);
    procedure fSetNOR(Value : integer);
    procedure fSetDOR(Value : integer);
    function fIsStop : boolean;
    function fProcess:double;
    procedure fSetMaxIteration(Value : integer);
  public
    //     
    property Delta            : double read fDelta;
    //   
    property Iteration        : integer read fIteration;
    //     
    function CalcDist(In1,In2:array of double):double;virtual;
    //  
    constructor Create(AOwner : TComponent);override;
    //    - 
    procedure Learn;
    //      
    function  Analize(In1:array of double):integer;
    //   - 
    procedure SaveToStream(S:TStream);
    //  
    procedure LoadFromStream(S:TStream);
  published
    //  
    property NumberOfClusters : integer read fNOC write fSetNOC;
    //  
    property NumberOfRecords  : integer read fNOR write fSetNOR;
    //  
    property DegreeOfRecord   : integer read fDOR write fSetDOR;
    //   
    property MaxIteration     : integer read fMaxIteration write fSetMaxIteration;
    //   
    property GetRecordProc    : GetRecordProc read fGetRecordProc write fGetRecordProc;
    //  -  
    property OnAfterIteration : TNotifyEvent read fAfterEpochProc write fAfterEpochProc;
  end;

procedure Register;

implementation
{ TCMeansClustering}
{ private }
//  
procedure TCMeansClustering.fInit;
var
  xI    : integer;
begin
  //   
  SetLength(fV,fNOC);
  SetLength(fW,fNOC);
  for xI:=0 to fNOC-1 do
  begin
    SetLength(fV[xI],fDOR);
    fW[xI] := 1;
  end;
  //   
  SetLength(fValues,fDOR);
  SetLength(fClustersW,fNOC);
  for xI := 0 to 1 do
    SetLength(fStopCheck[xI],fNOR);
end;
//   
procedure TCMeansClustering.fSetNOC;
begin
  if Value < 1 then
    raise Exception.Create('Number of clusters must be over 0');
  fNOC := Value;
  fInit;
end;
//    
procedure TCMeansClustering.fSetNOR;
begin
  if Value < 1 then
    raise Exception.Create('Number of records must be over 0');
  fNOR := Value;
  fInit;
end;
//   
procedure TCMeansClustering.fSetDOR;
begin
  if Value < 1 then
    raise Exception.Create('Degree of records must be over 0');
  fDOR := Value;
  fInit;
end;
//    
procedure TCMeansClustering.fSetMaxIteration;
begin
  if Value < 1 then
    raise Exception.Create('Max Iteration must be over 0');
  fMaxIteration := Value;  
end;
//   
function TCMeansClustering.fProcess;
var
  xMinDst   : double;
  xMinIndex : integer;
  xJ        : integer;
  xDst      : double;
begin
  xMinDst := CalcDist(fV[0],fValues);
  xMinIndex := 0;
  for xJ:=1 to fNOC-1 do
  begin
    xDst := CalcDist(fV[xJ],fValues);
    if xDst < xMinDst then
    begin
      xMinDst := xDst;
      xMinIndex := xJ;
    end;
  end;
  //   
  result := 0;
  if (xMinIndex<0) or (xMinIndex>fNOC-1) then raise Exception.Create('...');
  fNum := xMinIndex;
  for xJ := 0 to fDOR-1 do
  begin
    xDst := (fW[xMinIndex]*fV[xMinIndex,xJ]+fValues[xJ])/(fW[xMinIndex]+1);
    Result := Result + Abs(xDst-fV[xMinIndex,xJ]);
    fV[xMinIndex,xJ] := xDst;
  end;
  inc(fW[xMinIndex]);
end;
//      ,     - 
function TCMeansClustering.fIsStop;
var
  xI,xJ : integer;
  xNum  : integer;
begin
  xNum := fIteration mod 2;
  Result := True;
  for xI:=0 to fNOC-1 do
  begin
    fClustersW[xI] := 0;
    for xJ:=0 to fDOR-1 do
      fV[xI,xJ] := 0;
  end;
  for xI:=0 to fNOR-1 do
  begin
    fGetRecordProc(xI,fValues);
    if fStopCheck[0,xI] <> fStopCheck[1,xI] then
      Result := False;
    //   
    for xJ:=0 to fDOR-1 do
      fV[fStopCheck[xNum,xI],xJ] := fV[fStopCheck[xNum,xI],xJ] + fValues[xJ];
    inc(fClustersW[fStopCheck[xNum,xI]])
  end;
  //    
  for xI:=0 to fNOC-1 do
    if fClustersW[xI] > 0 then
      for xJ:=0 to fDOR-1 do
        fV[xI,xJ] := fV[xI,xJ]/fClustersW[xI];
end;

constructor TCMeansClustering.Create;
begin
  inherited;
  fNOC := 1;
  fNOR := 1;
  fDOR := 1;
  fMaxIteration := 1;
  fInit;
end;
//         
//        
function TCMeansClustering.CalcDist;
var
  xI : integer;
  xDst : double;
begin
  xDst := 0;
  for xI:=0 to fDOR-1 do
  begin
    xDst := xDst + sqr(in1[xI]-in2[xI]);
  end;
  Result := xDst;
end;
//   - 
procedure TCMeansClustering.Learn;
var
  xI,xJ : integer;
  xNum  : integer;
begin
  //  
  if fNOC > fNOR then
    raise Exception.Create('Number of records must be over number of clusters');
  fIteration := 0;
  xNum := fIteration mod 2;
  //    (-)
  for xI := 0 to fNOC-1 do
  begin
    fGetRecordProc(xI{Random(fNOR)},fValues);
    for xJ := 0 to fDOR-1 do
    begin
      fV[xI,xJ] := fValues[xJ];
      fStopCheck[xNum,xJ] := xJ;
    end;
  end;
  fDelta := 0;
  //     
  for xI := fNOC to fNOR-1 do
  begin
    //  
    fGetRecordProc(xI,fValues);
    //      
    fDelta := fDelta + fProcess;
    fStopCheck[xNum,xI] := fNum;
  end;
  if Assigned(OnAfterIteration) then
    OnAfterIteration(Self);
  //    
  for xI := 1 to fMaxIteration-1 do
  begin
    fIteration := xI;
    xNum := fIteration mod 2;
    fDelta := 0;
    for xJ := 0 to fNOR-1 do
    begin
      //  
      fGetRecordProc(xJ,fValues);
      //      
      fDelta := fDelta + fProcess;
      fStopCheck[xNum,xJ] := fNum;
    end;
    if Assigned(OnAfterIteration) then
      OnAfterIteration(Self);
    if fIsStop then break;
  end;
end;
//  ,    
function TCMeansClustering.Analize;
var
  xI : integer;
  xMinDst : double;
  xMinNum : integer;
  xDst    : double;
begin
  //     
  xMinDst := CalcDist(fV[0],in1);
  xMinNum := 0;
  for xI:=1 to fNOC-1 do
  begin
    xDst := CalcDist(fV[xI],in1);
    if xDst < xMinDst then
    begin
      xMinDst := xDst;
      xMinNum := xI;
    end;
  end;
  Result := xMinNum;
end;
procedure TCMeansClustering.SaveToStream;
var
  xI ,xJ: integer;
begin
  S.Write(fDOR,sizeof(fDOR));
  S.Write(fMaxIteration,sizeof(fMaxIteration));
  S.Write(fNOC,sizeof(fNOC));
  S.Write(fNOR,sizeof(fNOR));
  S.Write(fNum,sizeof(fNum));
  for xI := 0 to fNOC-1 do
  begin
    S.Write(fClustersW[xI],sizeof(fClustersW[xI]));
    S.Write(fW[xI],sizeof(fW[xI]));
    for xJ := 0 to fDOR-1 do
      S.Write(fV[xI,xJ],sizeof(fV[xI,xJ]));
  end;
end;
procedure TCMeansClustering.LoadFromStream;
var
  xI ,xJ: integer;
begin
  S.Read(fDOR,sizeof(fDOR));
  S.Read(fMaxIteration,sizeof(fMaxIteration));
  S.Read(fNOC,sizeof(fNOC));
  S.Read(fNOR,sizeof(fNOR));
  S.Read(fNum,sizeof(fNum));
  fInit;
  for xI := 0 to fNOC-1 do
  begin
    S.Read(fClustersW[xI],sizeof(fClustersW[xI]));
    S.Read(fW[xI],sizeof(fW[xI]));
    for xJ := 0 to fDOR-1 do
      S.Read(fV[xI,xJ],sizeof(fV[xI,xJ]));
  end;
end;

procedure Register;
begin
//  RegisterComponents('Statistics',[TCMeansClustering]);
end;

end.
